#include "kaestchen.h"
#include "spiellabel.h"
#include <QPainter>
#include <iostream>//TODO

using namespace std;

Kaestchen::Kaestchen(SpielLabel *spielLabel, int x, int y,
                     Kaestchen *linkerNachbar, Kaestchen *obererNachbar):
    spielLabel {spielLabel}, x {x}, y {y}
{
    this->linkerNachbar = linkerNachbar;
    this->obererNachbar = obererNachbar;
    if (obererNachbar) {
        seiteOben = obererNachbar->seiteUnten;
        seiteOben->setzeGeteilt();
        obererNachbar->untererNachbar = this;
    } else {
        seiteOben = {new Seite{spielLabel, x, y, Position::OBEN}};
    }
    seiteRechts = {new Seite{spielLabel, x, y, Position::RECHTS}};
    seiteUnten = {new Seite{spielLabel, x, y, Position::UNTEN}};
    if (linkerNachbar) {
        seiteLinks = linkerNachbar->seiteRechts;
        seiteLinks->setzeGeteilt();
        linkerNachbar->rechterNachbar = this;
    } else {
        seiteLinks = {new Seite{spielLabel, x, y, Position::LINKS}};
    }
    seiteOben->addNachbar(seiteRechts);
    seiteOben->addNachbar(seiteLinks);
    seiteUnten->addNachbar(seiteRechts);
    seiteUnten->addNachbar(seiteLinks);
}

Kaestchen::~Kaestchen()
{
    delete seiteOben;
    delete seiteRechts;
    delete seiteUnten;
    delete seiteLinks;
}

Kaestchen* Kaestchen::getNachbar(Position pos) const
{
    switch (pos) {
    case Position::OBEN: return obererNachbar;
    case Position::RECHTS: return rechterNachbar;
    case Position::UNTEN: return untererNachbar;
    default: return linkerNachbar;
    }
}

/*
 *Falls sich die Seite nicht markieren lässt,
 *weil sie z. B. zum Rand gehört oder schon markiert ist,
 *wird die Seite zurückgegeben. Ansonsten wird nullptr zurückgegeben.
 */
bool Kaestchen::markiereSeite(int x, int y, Position* seitenPosition)
{
    Seite* seite;
    if (x < y) {
        if (breite - x < y) {
            seite = seiteUnten;
            *seitenPosition = Position::UNTEN;
        } else {
            seite = seiteLinks;
            *seitenPosition = Position::LINKS;
        }
    } else {
        if (breite - x < y) {
            seite = seiteRechts;
            *seitenPosition = Position::RECHTS;
        } else {
            seite = seiteOben;
            *seitenPosition = Position::OBEN;
        }
    }
    return seite->setzeMarkiert();
}

void Kaestchen::markiereSeite(Position pos)
{
    Seite* seite;
    switch (pos) {
    case Position::OBEN: seite = seiteOben; break;
    case Position::RECHTS: seite = seiteRechts; break;
    case Position::UNTEN: seite = seiteUnten; break;
    default: seite = seiteLinks;
    }
    seite->setzeMarkiert();
}

Seite* Kaestchen::getSeite(Position pos) const
{
    switch (pos) {
    case Position::OBEN: return seiteOben;
    case Position::RECHTS: return seiteRechts;
    case Position::UNTEN: return seiteUnten;
    default: return seiteLinks;
    }
}

vector<Position> Kaestchen::getOffenePositionen() const
{
    vector<Position> positionen;
    if (seiteOben->istKlickbar())
        positionen.push_back(Position::OBEN);
    if (seiteRechts->istKlickbar())
        positionen.push_back(Position::RECHTS);
    if (seiteUnten->istKlickbar())
        positionen.push_back(Position::UNTEN);
    if (seiteLinks->istKlickbar())
        positionen.push_back(Position::LINKS);
    return positionen;
}

vector<Seite*> Kaestchen::getMarkierteSeiten() const
{
    vector<Seite*> seiten;
    if (!seiteOben->istKlickbar())
        seiten.push_back(seiteOben);
    if (!seiteRechts->istKlickbar())
        seiten.push_back(seiteRechts);
    if (!seiteUnten->istKlickbar())
        seiten.push_back(seiteUnten);
    if (!seiteLinks->istKlickbar())
        seiten.push_back(seiteLinks);
    return seiten;
}

bool Kaestchen::geschlossen() const
{
    return !seiteOben->istKlickbar() && !seiteRechts->istKlickbar()
            && !seiteUnten->istKlickbar() && !seiteLinks->istKlickbar();
}

void Kaestchen::zeichneGeschlossen(bool mensch)
{
    QPainter painter {spielLabel->getSpielflaeche()};
    QPen pen;
    pen.setWidth(2);
    painter.setPen(pen);
    int d = 0.1*breite;
    if (mensch) {
        painter.drawEllipse(x + d, y + d, breite - 2*d, breite - 2*d);
        //Mund
        pen.setColor(Qt::GlobalColor::red);
        painter.setPen(pen);
        painter.drawArc(x + 3*d, y + 5*d, 4*d, 2*d, -30*16, -120*16);
        //Augen
        pen.setWidth(3);
        pen.setColor(Qt::GlobalColor::blue);
        painter.setPen(pen);
        painter.drawEllipse(x + 3*d, y + 3*d, d, d);
        painter.drawEllipse(x + 6*d, y + 3*d, d, d);
        vomMenschenGeschlossen = true;
    } else {
        painter.drawRoundedRect(x + d, y + 2*d, breite - 2*d, breite - 4*d, 3, 3);
    }
    spielLabel->zeichne();
    if (mensch) spielLabel->incStandMensch(true);
    else spielLabel->incStandComputer(true);
}

void Kaestchen::graueAus()
{
    QPainter painter {spielLabel->getSpielflaeche()};
    painter.fillRect(x, y, breite, breite, grau);
    for (Seite* seite: {seiteOben, seiteRechts, seiteUnten, seiteLinks})
        seite->setzeNichtKlickbar();
    ausgegraut = true;
}

bool Kaestchen::istAusgegraut() const
{
    return ausgegraut;
}

void Kaestchen::setzeZurueck(bool mensch)
{
    QPainter painter {spielLabel->getSpielflaeche()};
    painter.fillRect(x + 2, y + 2, breite - 4, breite - 4,
                     SpielLabel::hintergrundfarbe);
    spielLabel->zeichne();
    if (mensch) spielLabel->incStandMensch(false);
    else spielLabel->incStandComputer(false);
    vomMenschenGeschlossen = false;
}

int Kaestchen::anzahlMarkierteSeiten() const
{
    return (seiteOben->istKlickbar() ? 0 : 1) +
            (seiteRechts->istKlickbar() ? 0 : 1) +
            (seiteUnten->istKlickbar() ? 0 : 1) +
            (seiteLinks->istKlickbar() ? 0 : 1);
}

char Kaestchen::status() const
{
    return seiteOben->istMarkiert()
            | (seiteRechts->istMarkiert() << 1)
            | (seiteUnten->istMarkiert() << 2)
            | (seiteLinks->istMarkiert() << 3)
            | (geschlossen() << 4)
            | (vomMenschenGeschlossen << 5)
            | (ausgegraut << 6);
}

void Kaestchen::setzeVomStatus(char status)
{
    if (status & 1)
        markiereSeite(Position::OBEN);
    if (status & 1 << 1)
        markiereSeite(Position::RECHTS);
    if (status & 1 << 2)
        markiereSeite(Position::UNTEN);
    if (status & 1 << 3)
        markiereSeite(Position::LINKS);
    if (static_cast<bool>(status & 1 << 4)) {
        vomMenschenGeschlossen = static_cast<bool>(status & 1 << 5);
        zeichneGeschlossen(vomMenschenGeschlossen);
    }
    if (status & 1 << 6)
        graueAus();
}
