#include "spiellabel.h"
#include "kaestchen.h"
#include <ctime>
#include <iostream>//TODO
#include <QFileDialog>
#include <QMessageBox>

using namespace std;

SpielLabel::SpielLabel(MainWindow* mainWindow)
    : QLabel {mainWindow},
      mainWindow {mainWindow}
{
    erzeugeSpielflaeche(Groesse::KLEIN, true);
    srand(time(nullptr));
    timerBlinke.setSingleShot(true);
    timerBlinke.setInterval(100);
    timerComputerIstDran.setSingleShot(true);
    timerComputerIstDran.setInterval(300);
    QObject::connect(&timerBlinke, SIGNAL(timeout()), this, SLOT(blinkeMitSeite()));
    QObject::connect(&timerComputerIstDran, SIGNAL(timeout()), this, SLOT(computerZug()));
}

void SpielLabel::erzeugeSpielflaeche(Groesse groesse, bool graueAus)
{
    timerBlinke.stop();
    timerComputerIstDran.stop();
    blinkZaehler = 0;
    kaestchenZumSchliessen.clear();
    int anzahlSpalten, anzahlZeilen;
    switch (groesse) {
    case Groesse::KLEIN:
        anzahlZeilen = 5;
        anzahlSpalten = 6;
        break;
    case Groesse::MITTEL:
        anzahlZeilen = 11;
        anzahlSpalten = 13;
        break;
    default:
        anzahlZeilen = 15;
        anzahlSpalten = 22;
    }
    anzahlKaestchen = anzahlZeilen*anzahlSpalten;
    b = anzahlSpalten*Kaestchen::breite + 2*abstandLinks + 1;
    h = anzahlZeilen*Kaestchen::breite + 2*abstandOben;
    resize(b, h);
    if (spielflaeche)
        delete spielflaeche;
    spielflaeche = new QPixmap {b, h};
    spielflaeche->fill(hintergrundfarbe);
    inKonstruktion = true;
    matrix = std::move(Matrix{anzahlZeilen, anzahlSpalten});
    for (int i {0}; i != anzahlZeilen; ++i) {
        for (int j {0}; j != anzahlSpalten; ++j) {
            matrix[i][j] = new Kaestchen{this,
                    abstandLinks + j*Kaestchen::breite, abstandOben + i*Kaestchen::breite,
                    j == 0 ? nullptr : matrix[i][j - 1],
                    i == 0 ? nullptr : matrix[i - 1][j]};
        }
    }
    if (graueAus && groesse == Groesse::MITTEL) {
        //Zeichne ausgegraute Winkel
        int teilpunktZeilen {3 + rand() % (anzahlZeilen - 6)};
        int teilpunktSpalten {3 + rand() % (anzahlSpalten - 6)};
        for (int i {0}; i != 4; ++i)
            anzahlKaestchen -=
                    zeichneWinkel((Ecke) (rand()%4),
                                  i < 2 ? 0 : teilpunktZeilen,
                                  i%2 ? 0 : teilpunktSpalten,
                                  (i < 2 ? teilpunktZeilen : anzahlZeilen) + (i < 2 ? 1: 0),
                                  (i%2 ? teilpunktSpalten : anzahlSpalten) + (i%2 ? 1: 0));
    }
    if (graueAus && groesse == Groesse::GROSS) {
        //Zeichne ausgegraute Winkel und Diagonalen
        int teilpunktZeilen {3 + rand() % (anzahlZeilen - 6)};
        int teilpunktSpalten1 {3 + rand() % (anzahlSpalten/2 - 6)};
        int teilpunktSpalten2 {6 + teilpunktSpalten1 + rand() % (anzahlSpalten/2 - 5)};
        for (int i {0}; i != 6; ++i)
            anzahlKaestchen -=
                    zeichneDiagonale(rand() & 1,
                                     i < 3 ? 1 : teilpunktZeilen + 1,
                                     (i%3 == 0 ? 0 : (i%3 == 1 ? teilpunktSpalten1 : teilpunktSpalten2)) + 1,
                                     i < 3 ? teilpunktZeilen - 1 : anzahlZeilen - 1,
                                     (i%3 == 0 ? teilpunktSpalten1 : (i%3 == 1 ? teilpunktSpalten2 : anzahlSpalten - 1)) - 1);
    }
    inKonstruktion = false;
    zeichne();
    standMensch = standComputer = 0;
    fertig = false;
    wechsleSpieler(menschStarted);
    zuege.clear();
    zugIt = zuege.begin();
    mainWindow->erlaubeRueckgaengig(false);
    mainWindow->erlaubeWiederherstellen(false);
    if (!menschIstDran) timerComputerIstDran.start();
}

int SpielLabel::zeichneWinkel(Ecke aussparung,
                              int minZeile, int minSpalte,
                              int maxZeile, int maxSpalte)
{
    int z0 {minZeile + 1 + rand() % (maxZeile - minZeile - 3)};
    int z1 {z0 + 1 + rand() % (maxZeile - z0 - 2)};
    int z2 {z1 + 1 + rand() % (maxZeile - z1 - 1)};
    int s0 {minSpalte + 1 + rand() % (maxSpalte - minSpalte - 3)};
    int s1 {s0 + 1 + rand() % (maxSpalte - s0 - 2)};
    int s2 {s1 + 1 + rand() % (maxSpalte - s1 - 1)};
    int res {0};
    if (aussparung != Ecke::LINKS_OBEN)
        for (int i {z0}; i != z1; ++i)
            for (int j {s0}; j != s1; ++j) {
                matrix[i][j]->graueAus();
                ++res;
            }
    if (aussparung != Ecke::RECHTS_OBEN)
        for (int i {z0}; i != z1; ++i)
            for (int j {s1}; j != s2; ++j) {
                matrix[i][j]->graueAus();
                ++res;
            }
    if (aussparung != Ecke::RECHTS_UNTEN)
        for (int i {z1}; i != z2; ++i)
            for (int j {s1}; j != s2; ++j) {
                matrix[i][j]->graueAus();
                ++res;
            }
    if (aussparung != Ecke::LINKS_UNTEN)
        for (int i {z1}; i != z2; ++i)
            for (int j {s0}; j != s1; ++j) {
                matrix[i][j]->graueAus();
                ++res;
            }
    return res;
}

int SpielLabel::zeichneDiagonale(bool vonLinksObenNachRechtsUnten,
                                 int minZeile, int minSpalte,
                                 int maxZeile, int maxSpalte)
{
    int res {0};
    int zDiff {maxZeile - minZeile};
    int sDiff {maxSpalte - minSpalte};
    bool iteriereUeberZeilen {zDiff >= sDiff};
    int i0 = iteriereUeberZeilen ? minZeile : minSpalte;
    int i1 = iteriereUeberZeilen ? maxZeile : maxSpalte;
    int j0 = iteriereUeberZeilen ? minSpalte : minZeile;
    double q = 1.*(iteriereUeberZeilen ? sDiff : zDiff)/(iteriereUeberZeilen ? zDiff : sDiff);
    double c = q*i0;
    for (int i {i0}; i != i1; ++i) {
        int s = iteriereUeberZeilen ? (j0 + (int) (q*i - c)) : i;
        if (!vonLinksObenNachRechtsUnten)
            s = maxSpalte + minSpalte - s;
        matrix[iteriereUeberZeilen ? i : (j0 + (int) (q*i - c))][s]->graueAus();
        ++res;
    }
    return res;
}

void SpielLabel::aktualisiereStand()
{
    fertig = standMensch + standComputer == anzahlKaestchen;
    QString s {"Mensch: "};
    s += to_string(standMensch).c_str();
    s += ", Computer: ";
    s += to_string(standComputer).c_str();
    if (!fertig && menschIstDran)
        s += ", du bist dran";
    if (fertig)
        s += ", fertig";
    mainWindow->aktualisiereStatusBar(s);
}

void SpielLabel::incStandMensch(bool inc)
{
    if (inc) ++standMensch;
    else --standMensch;
    aktualisiereStand();
}

void SpielLabel::incStandComputer(bool inc)
{
    if (inc) ++standComputer;
    else --standComputer;
    aktualisiereStand();
}

void SpielLabel::wechsleSpieler(bool menschIstDran)
{
    this->menschIstDran = menschIstDran;
    aktualisiereStand();
}

QPixmap* SpielLabel::getSpielflaeche() const
{
    return spielflaeche;
}

void SpielLabel::zeichne()
{
    if (!inKonstruktion) setPixmap(*spielflaeche);
}

int SpielLabel::breite() const
{
    return b;
}

int SpielLabel::hoehe() const
{
    return h;
}

void SpielLabel::mousePressEvent(QMouseEvent* event)
{
    if (!menschIstDran || event->button() != Qt::LeftButton
            || blinkZaehler || fertig)
        return;
    if (!zuege.empty() && (*zugIt) != zuege.back()) {
        mainWindow->erlaubeWiederherstellen(false);
        ++zugIt;
        for (list<Zug*>::iterator it {zugIt}; it != zuege.end();) {
            delete *it;
            it = zuege.erase(it);
        }
        zugIt = zuege.end();
        --zugIt;
    }
    int nx = (event->x() - abstandLinks)/Kaestchen::breite;
    int ny = (event->y() - abstandOben)/Kaestchen::breite;
    int kx = event->x() - abstandLinks - nx*Kaestchen::breite;
    int ky = event->y() - abstandOben - ny*Kaestchen::breite;
    Kaestchen* kaestchen = matrix[ny][nx];
    if (kaestchen->istAusgegraut()) return;
    Position pos;
    if (!kaestchen->markiereSeite(kx, ky, &pos)) {
        blinkSeite = kaestchen->getSeite(pos);
        blinkSeite->blinke(true);
        blinkZaehler = 1;
        blinkeMitWarnung = true;
        timerBlinke.start();
        return;
    }
    if (zuege.empty() || menschIstDran != (*zugIt)->mensch)
        neuerZug(new Zug {menschIstDran, kaestchen->getSeite(pos)});
    else
        (*zugIt)->seiten.push_back(kaestchen->getSeite(pos));
    bool kaestchenGeschlossen {kaestchen->geschlossen()};
    if (kaestchenGeschlossen) {
        kaestchen->zeichneGeschlossen(true);
        (*zugIt)->geschlosseneKaestchen.push_back(kaestchen);
    }
    Kaestchen* benachbartesKaestchen = kaestchen->getNachbar(pos);
    bool benachbartesKaestchenGeschlossen {benachbartesKaestchen->geschlossen()};
    if (benachbartesKaestchenGeschlossen) {
        benachbartesKaestchen->zeichneGeschlossen(true);
        (*zugIt)->geschlosseneKaestchen.push_back(benachbartesKaestchen);
    }
    if (kaestchenGeschlossen || benachbartesKaestchenGeschlossen)
        return;
    wechsleSpieler(false);
    timerComputerIstDran.start();
}

/*
 *Führt einen Zug mit dem Computer aus.
 *
 *- Gibt es ein Kästchen, bei dem drei Seiten markiert sind?
 *  Dann markiere die letzte Seite.
 *- Gibt es ein Kästchen, bei dem höchstens eine Seite markiert ist?
 *  Dann markiere eine beliebige Seite zu einem Nachbarkästchen,
 *  bei dem höchstens eine Seite markiert ist.
 *- Wähle ein Kästchen aus, bei dem zwei Seiten markiert sind,
 *  und markiere eine beliebige Seite,
 *  so dass der Mensch möglichst wenig Kästchen schließen kann.
 *
 */
void SpielLabel::computerZug()
{
    kaestchenMit3MarkiertenSeiten.clear();
    kaestchenMitMax1MarkierterSeite.clear();
    kaestchenMit2MarkiertenSeiten.clear();
    for (Kaestchen* kaestchen: matrix) {
        if (kaestchen->anzahlMarkierteSeiten() == 3)
            kaestchenMit3MarkiertenSeiten.push_back(kaestchen);
        else if (kaestchen->anzahlMarkierteSeiten() <= 1)
            kaestchenMitMax1MarkierterSeite.push_back(kaestchen);
        else if (kaestchen->anzahlMarkierteSeiten() == 2)
            kaestchenMit2MarkiertenSeiten.push_back(kaestchen);
    }
    if (rand() % 5 >= level
            && (!kaestchenMit2MarkiertenSeiten.empty() || !kaestchenMitMax1MarkierterSeite.empty()))
        kaestchenMit3MarkiertenSeiten.clear();
    if (rand() % 5 >= level
            && kaestchenMit3MarkiertenSeiten.empty() && !kaestchenMit2MarkiertenSeiten.empty())
        kaestchenMitMax1MarkierterSeite.clear();
    blinkSeite = nullptr;
    if (!kaestchenMit3MarkiertenSeiten.empty()) {
        Kaestchen* kaestchen {kaestchenMit3MarkiertenSeiten[rand() % kaestchenMit3MarkiertenSeiten.size()]};
        vector<Position> positionen {kaestchen->getOffenePositionen()};
        Position pos {positionen[0]};
        blinkSeite = kaestchen->getSeite(pos);
        kaestchenZumSchliessen.push_back(kaestchen);
        Kaestchen* nachbar {kaestchen->getNachbar(pos)};
        if (nachbar->anzahlMarkierteSeiten() == 3)
            kaestchenZumSchliessen.push_back(nachbar);
        kaestchen->markiereSeite(pos);
    }
    if (!blinkSeite && !kaestchenMitMax1MarkierterSeite.empty()) {
        do {
            int i = rand() % kaestchenMitMax1MarkierterSeite.size();
            Kaestchen* kaestchen = kaestchenMitMax1MarkierterSeite[i];
            vector<Position> positionen = {Position::OBEN, Position::RECHTS, Position::UNTEN, Position::LINKS};
            do {
                int j = rand() % positionen.size();
                Position pos {positionen[j]};
                Kaestchen* nachbar = kaestchen->getNachbar(pos);
                if (nachbar && nachbar->anzahlMarkierteSeiten() <= 1
                        && (nachbar->anzahlMarkierteSeiten() == 0
                            || nachbar->getMarkierteSeiten()[0] != kaestchen->getSeite(pos))
                        ) {
                    kaestchen->markiereSeite(pos);
                    blinkSeite = kaestchen->getSeite(pos);
                    break;
                }
                positionen.erase(positionen.begin() + j);
            } while (!positionen.empty());
            if (blinkSeite) break;
            kaestchenMitMax1MarkierterSeite.erase(kaestchenMitMax1MarkierterSeite.begin() + i);
        } while (!kaestchenMitMax1MarkierterSeite.empty());
    }
    if (!blinkSeite && !kaestchenMit2MarkiertenSeiten.empty()) {
        unsigned int anzahlSchliessbareKaestchen = 0xffffff;
        Kaestchen* gefundenesKaestchen;
        /*Pro Schleifendurchlauf werden alle Kästchen gefunden,
         *die der Mensch im darauffolgenden Zug schließen kann,
         *falls kaestchen markiert wird.
         */
        do {
            Kaestchen* kaestchen {kaestchenMit2MarkiertenSeiten[rand() % kaestchenMit2MarkiertenSeiten.size()]};
            if (rand() % 5 >= level) {
                gefundenesKaestchen = kaestchen;
                break;
            }
            vector<Kaestchen*> schliessbareKaestchen;
            untersucheKaestchen(kaestchen, schliessbareKaestchen);
            if (schliessbareKaestchen.size() < anzahlSchliessbareKaestchen) {
                anzahlSchliessbareKaestchen = schliessbareKaestchen.size();
                gefundenesKaestchen = kaestchen;
            }
        } while (!kaestchenMit2MarkiertenSeiten.empty());
        vector<Position> positionen {gefundenesKaestchen->getOffenePositionen()};
        Position pos {positionen[rand() % positionen.size()]};
        Kaestchen* nachbar {gefundenesKaestchen->getNachbar(pos)};
        if (nachbar->anzahlMarkierteSeiten() == 3)
            kaestchenZumSchliessen.push_back(nachbar);
        gefundenesKaestchen->markiereSeite(pos);
        blinkSeite = gefundenesKaestchen->getSeite(pos);
    }
    if (menschIstDran != (*zugIt)->mensch)
        neuerZug(new Zug {menschIstDran, blinkSeite});
    else
        (*zugIt)->seiten.push_back(blinkSeite);
    blinkZaehler = 2;
    blinkeMitWarnung = false;
    timerBlinke.start();
}

static bool enthaelt(vector<Kaestchen*>& v, Kaestchen* kaestchen)
{
    return find(v.begin(), v.end(), kaestchen) != v.end();
}

static void loesche(vector<Kaestchen*>& v, Kaestchen* kaestchen)
{
    v.erase(find(v.begin(), v.end(), kaestchen));
}

void SpielLabel::untersucheKaestchen(Kaestchen *kaestchen, vector<Kaestchen*>& schliessbareKaestchen)
{
    if (kaestchen->anzahlMarkierteSeiten() == 2) {
        if (enthaelt(kaestchenMit2MarkiertenSeiten, kaestchen))
            loesche(kaestchenMit2MarkiertenSeiten, kaestchen);
        schliessbareKaestchen.push_back(kaestchen);
    } else {
        int anzahlKlickbareSeiten {4 - kaestchen->anzahlMarkierteSeiten()};
        for (Position pos: kaestchen->getOffenePositionen()) {
            if (enthaelt(schliessbareKaestchen, kaestchen->getNachbar(pos)))
                --anzahlKlickbareSeiten;
        }
        if (anzahlKlickbareSeiten <= 1)
            schliessbareKaestchen.push_back(kaestchen);
        else
            return;
    }
    for (Position pos: kaestchen->getOffenePositionen()) {
        Kaestchen* nachbar = kaestchen->getNachbar(pos);
        if (!enthaelt(schliessbareKaestchen, nachbar))
            untersucheKaestchen(nachbar, schliessbareKaestchen);
    }
}

void SpielLabel::blinkeMitSeite()
{
    if ((blinkZaehler & 1))
        blinkSeite->setzeNormal();
    else
        blinkSeite->blinke(blinkeMitWarnung);
    if (++blinkZaehler != 4)
        timerBlinke.start();
    else {
        blinkZaehler = 0;
        if (!menschIstDran) {
            if (!kaestchenZumSchliessen.empty()) {
                for (Kaestchen* kaestchen: kaestchenZumSchliessen) {
                    kaestchen->zeichneGeschlossen(false);
                    (*zugIt)->geschlosseneKaestchen.push_back(kaestchen);
                }
                kaestchenZumSchliessen.clear();
                if (!fertig) timerComputerIstDran.start();
            } else {
                wechsleSpieler(true);
            }
        }
    }
}

void SpielLabel::neuerZug(Zug* zug)
{
    zuege.push_back(zug);
    zugIt = zuege.end();
    --zugIt;
    mainWindow->erlaubeRueckgaengig(true);
}

void SpielLabel::rueckgaengig()
{
    if (!menschIstDran) return;
    mainWindow->erlaubeWiederherstellen(true);
    for (int i {0}; i != 2; ++i) {
        for (Seite* seite: (*zugIt)->seiten)
            seite->setzeZurueck();
        for (Kaestchen* kaestchen: (*zugIt)->geschlosseneKaestchen)
            kaestchen->setzeZurueck((*zugIt)->mensch);
        --zugIt;
    }
    if (++zugIt == zuege.begin())
        mainWindow->erlaubeRueckgaengig(false);
    --zugIt;
}

void SpielLabel::wiederherstellen()
{
    if (!menschIstDran) return;
    mainWindow->erlaubeRueckgaengig(true);
    for (int i {0}; i != 2; ++i) {
        ++zugIt;
        for (Seite* seite: (*zugIt)->seiten)
            seite->setzeMarkiert();
        for (Kaestchen* kaestchen: (*zugIt)->geschlosseneKaestchen)
            kaestchen->zeichneGeschlossen((*zugIt)->mensch);
    }
    if (++zugIt == zuege.end())
        mainWindow->erlaubeWiederherstellen(false);
    --zugIt;
}

void SpielLabel::setzeLevel(int level)
{
    this->level = level;
}

void SpielLabel::setzeMenschStarted(bool menschStarted)
{
    this->menschStarted = menschStarted;
}

static QString filter {"Käsekästchen (*.kk)"};

void SpielLabel::speichern(Groesse groesse, bool speichernUnter)
{
    if (speichernUnter || dateiName.isEmpty()) {
        dateiName = QFileDialog::getSaveFileName(
                    this,
                    QString {"Spiel speichern"} + (speichernUnter ? " unter ..." : ""),
                    dateiName.isEmpty() ? QDir::homePath() : dateiName,
                    filter);
    }
    if (dateiName.isEmpty()) return;
    QFile file {dateiName};
    file.open(QIODevice::WriteOnly);
    file.putChar(static_cast<char>(groesse));
    file.putChar(level);
    file.putChar(menschIstDran);
    bool res;
    for (Kaestchen* kaestchen: matrix)
        res = file.putChar(kaestchen->status());
    file.close();
    if (res)
        mainWindow->zeigeTemporaereStatusNachricht("Datei gespeichert");
    else
        QMessageBox::warning(this, "Fehler", "Die Datei kann nicht gespeichert werden.");
}

void SpielLabel::oeffnen()
{
    dateiName = QFileDialog::getOpenFileName(
                this, "Spiel öffnen",
                dateiName.isEmpty() ? QDir::homePath() : dateiName,
                filter);
    if (dateiName.isEmpty()) return;
    QFile file {dateiName};
    file.open(QIODevice::ReadOnly);
    char c;
    file.getChar(&c);
    if (c < 0 || c > 2) {
        c = static_cast<char>(Groesse::KLEIN);
    }
    mainWindow->setzeGroesse(static_cast<Groesse>(c), false, true);
    file.getChar(&c);
    mainWindow->setzeLevel(c);
    file.getChar(&c);
    menschIstDran = static_cast<bool>(c);
    for (Kaestchen* kaestchen: matrix) {
        file.getChar(&c);
        kaestchen->setzeVomStatus(c);
    }
    file.close();
    if (!menschIstDran)
        timerComputerIstDran.start();
}
