#ifndef SPIELLABEL_H
#define SPIELLABEL_H

#include "mainwindow.h"
#include "kaestchen.h"
#include "seite.h"
#include "matrix.h"
#include <QLabel>
#include <QMouseEvent>
#include <QTimer>
#include <QPixmap>
#include <vector>
#include <climits>

class Kaestchen;
class MainWindow;
struct Zug;

enum class Groesse {KLEIN, MITTEL, GROSS};

class SpielLabel : public QLabel
{
    Q_OBJECT

    MainWindow* mainWindow;
    QPixmap* spielflaeche {nullptr};
    int level {5};
    bool menschStarted {true};
    int anzahlKaestchen;
    bool fertig;
    Matrix matrix;
    static const int abstandLinks {1};
    static const int abstandOben {2};
    int b;
    int h;
    QTimer timerBlinke {this};
    QTimer timerComputerIstDran {this};
    Seite* blinkSeite;
    std::vector<Kaestchen*> kaestchenZumSchliessen;
    int blinkZaehler {0};
    bool blinkeMitWarnung;
    bool menschIstDran;
    int standMensch, standComputer;
    void aktualisiereStand();
    void wechsleSpieler(bool menschIstDran);
    bool inKonstruktion;
    std::vector<Kaestchen*> kaestchenMit3MarkiertenSeiten;
    std::vector<Kaestchen*> kaestchenMitMax1MarkierterSeite;
    std::vector<Kaestchen*> kaestchenMit2MarkiertenSeiten;
    void untersucheKaestchen(Kaestchen* kaestchen,
                             std::vector<Kaestchen*>& schliessbareKaestchen);
    struct Zug
    {
        bool mensch;
        std::vector<Seite*> seiten;
        std::vector<Kaestchen*> geschlosseneKaestchen;
        Zug(bool mensch, Seite* seite):
            mensch {mensch}, seiten {seite} {}
    };
    std::list<Zug*> zuege;
    std::list<Zug*>::iterator zugIt;
    void neuerZug(Zug* zug);
    enum class Ecke {LINKS_OBEN, RECHTS_OBEN,
                     RECHTS_UNTEN, LINKS_UNTEN};
    int zeichneWinkel(Ecke aussparung,
                      int minZeile, int minSpalte,
                      int maxZeile, int maxSpalte);
    int zeichneDiagonale(bool vonLinksObenNachRechtsUnten,
                         int minZeile, int minSpalte,
                         int maxZeile, int maxSpalte);
    QString dateiName;

private slots:
    void blinkeMitSeite();
    void computerZug();

public:
    SpielLabel(MainWindow *mainwindow);
    QPixmap* getSpielflaeche() const;
    void zeichne();
    int breite() const;
    int hoehe() const;
    void erzeugeSpielflaeche(Groesse groesse, bool graueAus);
    void setzeLevel(int level);
    void setzeMenschStarted(bool menschStarted);
    void incStandMensch(bool inc);
    void incStandComputer(bool inc);
    void rueckgaengig();
    void wiederherstellen();
    void speichern(Groesse groesse, bool speichernUnter);
    void oeffnen();
    static const Qt::GlobalColor hintergrundfarbe {Qt::GlobalColor::yellow};

protected:
    void mousePressEvent(QMouseEvent*);
};

#endif // SPIELLABEL_H
